import 'dart:async';

import 'package:dartz/dartz.dart';
import 'package:flutter/foundation.dart';
import 'package:hive/hive.dart';
import 'package:path_provider/path_provider.dart';
import 'package:synchronized/synchronized.dart';

abstract class Storage {
  Option<dynamic> read(String key);
  Future<bool> write(String key, dynamic value);
  Future<bool> delete(String key);
  Future<bool> clear();
}

class NullStorage implements Storage {
  @override
  Option<dynamic> read(String key) => none();

  @override
  Future<bool> write(String key, value) => Future.value(true);

  @override
  Future<bool> delete(String key) => Future.value(true);

  @override
  Future<bool> clear() => Future.value(true);
}

class HiveStorage implements Storage {
  HiveStorage._(this._box);

  static Future<HiveStorage> build() {
    return _lock.synchronized(
        () => _instance.map((i) => Future.value(i)).getOrElse(() async {
              final dir = await getTemporaryDirectory();
              if (!kIsWeb) Hive.init(dir.path);

              final box = await Hive.openBox<dynamic>('persisted_bloc_stream');

              final instance = HiveStorage._(box);
              _instance = some(instance);
              return instance;
            }));
  }

  static final _lock = Lock();
  static Option<HiveStorage> _instance = none();

  final Box<dynamic> _box;

  Future<bool> _withBox(void Function(Box<dynamic>) f) => _box.isOpen
      ? _lock.synchronized(() {
          f(_box);
          return true;
        })
      : Future.value(false);

  @override
  Option<dynamic> read(String key) =>
      _box.isOpen ? optionOf(_box.get(key)) : none();

  @override
  Future<bool> write(String key, dynamic value) =>
      _withBox((box) => box.put(key, value));

  @override
  Future<bool> delete(String key) => _withBox((box) => box.delete(key));

  @override
  Future<bool> clear() => _withBox((box) => box.clear());
}
